/* mpz_bc_bin_uiui - compute n over k.

Copyright 2011 Free Software Foundation, Inc.

This file is part of the GNU MP Library.

The GNU MP Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 3 of the License, or (at your
option) any later version.

The GNU MP Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the GNU MP Library.  If not, see http://www.gnu.org/licenses/.  */

#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"

/* Algorithm:

   Plain and simply multiply things together.

   We tabulate inverses (k/2^t)^(-1) mod B for 1= < k <= 32 (where t is chosen
   such that k/2^t is odd).
*/

/* To avoid generating this table, we could use constant expressions like
   (B-1) / 3 + 1.  */
static mp_limb_t zinv[] = {
#if GMP_NUMB_BITS == 64
  0x0000000000000001,	/*  1 */  0x0000000000000001,	/*  2 */
  0xaaaaaaaaaaaaaaab,	/*  3 */  0x0000000000000001,	/*  4 */
  0xcccccccccccccccd,	/*  5 */  0xaaaaaaaaaaaaaaab,	/*  6 */
  0x6db6db6db6db6db7,	/*  7 */  0x0000000000000001,	/*  8 */
  0x8e38e38e38e38e39,	/*  9 */  0xcccccccccccccccd,	/* 10 */
  0x2e8ba2e8ba2e8ba3,	/* 11 */  0xaaaaaaaaaaaaaaab,	/* 12 */
  0x4ec4ec4ec4ec4ec5,	/* 13 */  0x6db6db6db6db6db7,	/* 14 */
  0xeeeeeeeeeeeeeeef,	/* 15 */  0x0000000000000001,	/* 16 */
  0xf0f0f0f0f0f0f0f1,	/* 17 */  0x8e38e38e38e38e39,	/* 18 */
  0x86bca1af286bca1b,	/* 19 */  0xcccccccccccccccd,	/* 20 */
  0xcf3cf3cf3cf3cf3d,	/* 21 */  0x2e8ba2e8ba2e8ba3,	/* 22 */
  0xd37a6f4de9bd37a7,	/* 23 */  0xaaaaaaaaaaaaaaab,	/* 24 */
  0x8f5c28f5c28f5c29,	/* 25 */  0x4ec4ec4ec4ec4ec5,	/* 26 */
  0x84bda12f684bda13,	/* 27 */  0x6db6db6db6db6db7,	/* 28 */
  0x34f72c234f72c235,	/* 29 */  0xeeeeeeeeeeeeeeef,	/* 30 */
  0xef7bdef7bdef7bdf,	/* 31 */  0x0000000000000001	/* 32 */
#endif
#if GMP_NUMB_BITS == 32
  0x00000001,	/*  1 */  0x00000001,	/*  2 */
  0xaaaaaaab,	/*  3 */  0x00000001,	/*  4 */
  0xcccccccd,	/*  5 */  0xaaaaaaab,	/*  6 */
  0xb6db6db7,	/*  7 */  0x00000001,	/*  8 */
  0x38e38e39,	/*  9 */  0xcccccccd,	/* 10 */
  0xba2e8ba3,	/* 11 */  0xaaaaaaab,	/* 12 */
  0xc4ec4ec5,	/* 13 */  0xb6db6db7,	/* 14 */
  0xeeeeeeef,	/* 15 */  0x00000001,	/* 16 */
  0xf0f0f0f1,	/* 17 */  0x38e38e39,	/* 18 */
  0x286bca1b,	/* 19 */  0xcccccccd,	/* 20 */
  0x3cf3cf3d,	/* 21 */  0xba2e8ba3,	/* 22 */
  0xe9bd37a7,	/* 23 */  0xaaaaaaab,	/* 24 */
  0xc28f5c29,	/* 25 */  0xc4ec4ec5,	/* 26 */
  0x684bda13,	/* 27 */  0xb6db6db7,	/* 28 */
  0x4f72c235,	/* 29 */  0xeeeeeeef,	/* 30 */
  0xbdef7bdf,	/* 31 */  0x00000001	/* 32 */
#endif
};

static unsigned char ctz[] = {0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5};

void
mpz_bc_bin_uiui (mpz_ptr r, unsigned long int n, unsigned long int k)
{
  mp_limb_t i, j, c;

  c = n - k + 1;
  i = n - k + 2;
  for (j = 1; j < k; j++)
    {
      c = c * i;
      i++;
      c = c * zinv[j] >> ctz[j];
    }

  PTR(r)[0] = c;
  SIZ(r) = 1;
}

#ifdef TEST
#include <stdio.h>
#include <stdlib.h>

int
main (int argc, char **argv)
{
  unsigned long n, k;
  mpz_t r, ref;

  mpz_inits (r, ref, 0);

  for (k = 1; k < GMP_NUMB_BITS / 2; k++)
    {
      for (n = k; n <= GMP_LIMB_BITS + 2; n++)
	{
	  mpz_bin_uiui (ref, n, k);
	  mpz_bc_bin_uiui (r, n, k);
	  if (mpz_cmp (ref, r) != 0)
	    {
	      printf ("\r%ld over %ld fails", n, k);
	      abort ();
	    }
	}
    }

  return 0;
}
#endif

#ifdef MAIN
#include <stdlib.h>

int
main (int argc, char **argv)
{
  unsigned long n, k;
  mpz_t r;

  n = strtoul (argv[1], 0, 0);
  k = strtoul (argv[2], 0, 0);

  mpz_init (r);

  mpz_bc_bin_uiui (r, n, k);
  gmp_printf ("%Zx\n", r);

  return 0;
}
#endif
